#ifndef __KAN_ABSTRACTCOLLECTION_H__
#define __KAN_ABSTRACTCOLLECTION_H__

#include "kan/util.h"
#include "kan/mem.h"

template <typename T>
class AbstractCollection: public ICollection<T>
{
protected:
    AbstractCollection() {};
    
public:
    ~AbstractCollection() {};

    virtual boolean     add(T e) = 0;
    virtual boolean     remove(T e);
    virtual boolean     contains(T e);

    virtual boolean     addAll(ICollection<T> *c);
    virtual boolean     removeAll(ICollection<T> *c);
    virtual boolean     retainsAll(ICollection<T> *c);
    virtual boolean     containsAll(ICollection<T> *c);

    virtual void        clear();
    virtual size_t      size() = 0;
    virtual boolean     isEmpty(T e);

    virtual IIterator<T>*   iterator() = 0;
    virtual IComparator<T>*  comparator() = 0;
};

template <typename T>
boolean AbstractCollection<T>::isEmpty(T e) 
{
    return (size() == 0) ? uTrue : uFalse;
}

template <typename T>
boolean AbstractCollection<T>::contains(T e) 
{
    IIterator<T> *itr = iterator();
    IComparator<T> *comp = comparator();
    while (itr->hasNext()) { 
        if (comp->compare(itr->next(), e) == 0) { 
            return uTrue;
        }
    }
    return uFalse;
}

template <typename T>
boolean AbstractCollection<T>::remove(T e) 
{
    IIterator<T> *itr = iterator();
    IComparator<T> *comp = comparator();
    while (itr->hasNext()) { 
        if (comp->compare(itr->next(), e) == 0) { 
            itr->remove();
            return uTrue;
        }
    }
    return uFalse;
}

template <typename T>
boolean AbstractCollection<T>::containsAll(ICollection<T> *c) 
{
    if (c == NULL) return uFalse;

    IIterator<T> *itr = c.iterator();

    while (itr->hasNext()) { 
        if (contains(itr->next()) != uTrue) {
            return uFalse;
        }
    }
    return uTrue;
}

template <typename T>
boolean AbstractCollection<T>::addAll(ICollection<T> *c) 
{
    boolean modified = uFalse;
    if (c == NULL) return uFalse;

    IIterator<T> *itr = c.iterator();

    while (itr->hasNext()) { 
        if (add(itr->next()) == uTrue) {
            modified = uTrue;
        }
    }
    return modified;
}

template <typename T>
boolean AbstractCollection<T>::removeAll(ICollection<T> *c) 
{
    boolean modified = uFalse;
    if (c == NULL) return uFalse;

    IIterator<T> *itr = iterator();

    while (itr->hasNext()) { 
        if (c->contains(itr->next()) == uTrue) {
            itr->remove();
            modified = uTrue;
        }
    }
    return modified;
}

template <typename T>
boolean AbstractCollection<T>::retainsAll(ICollection<T> *c) 
{
    boolean modified = uFalse;
    if (c == NULL) return uFalse;

    IIterator<T> *itr = iterator();

    while (itr->hasNext()) { 
        if (c->contains(itr->next()) == uFalse) {
            itr->remove();
            modified = uTrue;
        }
    }
    return modified;
}

template <typename T>
void AbstractCollection<T>::clear()
{
    IIterator<T> *itr = iterator();

    while (itr->hasNext()) { 
        itr->next();
        itr->remove();
    }
    return ;
}

#endif  /* __KAN_ABSTRACTCOLLECTION_H__ */
